# Copyright 2023 Eurobios
# Licensed under the CeCILL License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     https://cecill.info/
from crep.base import (merge, aggregate_constant, aggregate_duplicates, merge_event, aggregate_on_segmentation)
from crep.tools import compute_discontinuity
